# -*- coding: utf-8 -*-

import unittest

from advanced_ssh_config.advanced_ssh_config import AdvancedSshConfig
from advanced_ssh_config.exceptions import ConfigError
from . import set_config, prepare_config, DEFAULT_CONFIG
from .. import __version__


class TestRoot(unittest.TestCase):
    def test_version(self):
        self.assertIsInstance(__version__, str)


class TestAdvancedSshConfig(unittest.TestCase):

    def setUp(self):
        prepare_config()

    def test_load_advanced_ssh_config(self):
        advssh = AdvancedSshConfig()
        self.assertIsInstance(advssh, AdvancedSshConfig)

    def test_routing_simple(self):
        advssh = AdvancedSshConfig(hostname='test',
                                   port=23,
                                   verbose=True,
                                   dry_run=True)
        routing = advssh.get_routing()
        self.assertEqual(routing['port'], 23)
        self.assertEqual(routing['hostname'], 'test')
        self.assertEqual(routing['reallocalcommand'], None)
        self.assertEqual(routing['gateways'], ['direct'])
        self.assertEqual(routing['verbose'], True)
        self.assertEqual(routing['proxy_type'], 'nc')
        self.assertEqual(routing['proxy_commands'][0],
                         ['nc', '-v', '-w', 180, '-G', 5, 'test', 23])

    def test_routing_hostname_in_config(self):
        contents = """
[test.com]
hostname = 1.2.3.4
port = 25
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test.com',
                                   port=25,
                                   verbose=True,
                                   dry_run=True,
                                   configfiles=[DEFAULT_CONFIG])
        routing = advssh.get_routing()
        self.assertEqual(routing['port'], 25)
        self.assertEqual(routing['hostname'], '1.2.3.4')
        self.assertEqual(routing['proxy_type'], 'nc')
        self.assertEqual(routing['proxy_commands'][0],
                         ['nc', '-v', '-w', 180, '-G', 5, '1.2.3.4', 25])

    def test_routing_config_override(self):
        contents = """
[test.com]
port = 25
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test.com',
                                   port=23,
                                   verbose=True,
                                   dry_run=True,
                                   configfiles=[DEFAULT_CONFIG])
        routing = advssh.get_routing()
        self.assertEqual(routing['port'], 23)
        self.assertEqual(routing['hostname'], 'test.com')
        self.assertEqual(routing['proxy_type'], 'nc')
        self.assertEqual(routing['proxy_commands'][0],
                         ['nc', '-v', '-w', 180, '-G', 5, 'test.com', 23])

    def test_routing_via_two_other_hosts(self):
        advssh = AdvancedSshConfig(hostname='aaa.com/bbb.com/ccc.com')
        routing = advssh.get_routing()
        self.assertEqual(routing['hostname'], 'aaa.com')
        self.assertEqual(routing['proxy_type'], 'nc')
        self.assertEqual(routing['gateways'], ['direct'])
        self.assertEqual(routing['proxy_commands'][0],
                         ['nc', '-w', 180, '-G', 5, 'aaa.com', 22])
        self.assertEqual(routing['gateway_route'], ['bbb.com', 'ccc.com'])

    def test_routing_via_two_other_hosts_with_config_one(self):
        contents = """
[ddd.com]
hostname = 1.2.3.4
port = 25
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='ddd.com/eee.com',
                                   configfiles=[DEFAULT_CONFIG])
        routing = advssh.get_routing()
        self.assertEqual(routing['hostname'], '1.2.3.4')
        self.assertEqual(routing['proxy_type'], 'nc')
        self.assertEqual(routing['gateways'], ['direct'])
        self.assertEqual(routing['proxy_commands'][0],
                         ['nc', '-w', 180, '-G', 5, '1.2.3.4', 25])
        self.assertEqual(routing['gateway_route'], ['eee.com'])

    def test_prepare_sshconfig_simple(self):
        contents = """
[test]
port = 25

[default]
port = 24
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        self.assertEqual(len(config.keys()), 2)
        self.assertEqual(config['test'].host, 'test')
        self.assertEqual(config['test'].config, [('port', '25')])
        self.assertEqual(config['default'].host, 'default')
        self.assertEqual(config['default'].config, [('port', '24')])

    def test_prepare_sshconfig_multiline(self):
        contents = """
[test]
localforward = 1 2.3.4.5 6 \n 7 8.9.10.11 12
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        self.assertEqual(config['test'].host, 'test')
        self.assertEqual(config['test'].config,
                         [('localforward', '1 2.3.4.5 6'),
                          ('localforward', '7 8.9.10.11 12')])

    def test_inherits(self):
        contents = """
[aaa]
hostname = 1.2.3.4
user = toto

[bbb]
inherits = aaa
port = 23
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full
        self.assertEqual(config['aaa'].clean_config['user'], ['toto'])
        self.assertEqual('port' in config['aaa'].clean_config, False)
        self.assertEqual(config['bbb'].clean_config['user'], ['toto'])
        self.assertEqual(config['bbb'].clean_config['port'], ['23'])

    def test_build_ssh_config(self):
        contents = """
[aaa]
hostname = 1.2.3.4
user = toto

[bbb]
inherits = aaa
port = 23
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        self.assertEquals(len(arr), 14)
        dest = """
# assh version: {}

Host aaa
  user toto
  # hostname 1.2.3.4

Host bbb
  port 23
  user toto
  # inherits aaa

Host *
  proxycommand assh connect %h --port=%p
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    def test_build_ssh_config_with_proxycommand(self):
        contents = """
[aaa]
hostname = 1.2.3.4
user = toto

[bbb]
inherits = aaa
port = 23

[default]
ProxyCommand = assh connect %h --port=%p
User = titi
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        #self.assertEquals(len(arr), 17)
        dest = """
# assh version: {}

Host aaa
  user toto
  # hostname 1.2.3.4

Host bbb
  port 23
  user toto
  # inherits aaa

Host *
  proxycommand assh connect %h --port=%p
  user titi
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    def test_build_ssh_config_sorted(self):
        contents = """
[ddd]
inherits = aaa
port = 23
user = titi

[bbb]
user = titi
inherits = aaa
port = 23
hostname = 1.1.1.1

[ccc]
hostname = 5.4.3.2
inherits = aaa
port = 23

[aaa]
hostname = 1.2.3.4
user = toto
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        dest = """
# assh version: {}

Host aaa
  user toto
  # hostname 1.2.3.4

Host bbb
  port 23
  user titi
  # hostname 1.1.1.1
  # inherits aaa

Host ccc
  port 23
  user toto
  # hostname 5.4.3.2
  # inherits aaa

Host ddd
  port 23
  user titi
  # inherits aaa

Host *
  proxycommand assh connect %h --port=%p
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    def test_inherits_noexists(self):
        contents = """
[aaa]
hostname = 1.2.3.4
user = toto

[bbb]
inherits = ccc
port = 23
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full

        def call():
            return config['bbb'].clean_config

        self.assertRaises(ConfigError, call)

    def test_inherits_deep(self):
        contents = """
[aaa]
hostname = 1.2.3.4
user = toto

[bbb]
inherits = aaa
tcpkeepalive = 42

[ccc]
inherits = bbb
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full
        self.assertEqual(config['ccc'].clean_config['user'], ['toto'])
        self.assertEqual(config['ccc'].clean_config['tcpkeepalive'], ['42'])

    def test_inherits_override(self):
        contents = """
[aaa]
user = toto

[bbb]
inherits = aaa
user = titi
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full
        self.assertEqual(config['aaa'].clean_config['user'], ['toto'])
        self.assertEqual(config['bbb'].clean_config['user'], ['titi'])

    def test_inherits_loop(self):
        contents = """
[aaa]
inherits = ccc

[bbb]
inherits = aaa

[ccc]
inherits = bbb
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full

        def call(key):
            return config[key].clean_config

        self.assertRaises(ConfigError, call, 'aaa')
        self.assertRaises(ConfigError, call, 'bbb')
        self.assertRaises(ConfigError, call, 'ccc')

    def test_inherits_loop_self(self):
        contents = """
[aaa]
inherits = aaa
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full

        def call(key):
            return config[key].clean_config

        self.assertRaises(ConfigError, call, 'aaa')

    def test_reserved_key(self):
        contents = """
[aaa]
user = toto
proxycommand = nc
hostname = titi
alias = tutu
gateways = toutou
reallocalcommand = tonton
remotecommand = tantan
includes = tuotuo
inherits = bbb
password = 4242
comment = 4343
[bbb]
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.config.full
        self.assertEquals(
            config['aaa'].clean_config,
            {'user': ['toto'], 'proxycommand': ['nc']}
        )

    def test_comment_simple(self):
        contents = """
[test]
comment = Hello
"""
        config = set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        routing = advssh.get_routing()
        self.assertEquals(routing['comment'], ['Hello'])

    def test_comment_multiline(self):
        contents = """
[test]
comment = Hello
          World
                         !

port = 22
"""
        config = set_config(contents)
        advssh = AdvancedSshConfig(hostname='test',
                                   configfiles=[DEFAULT_CONFIG])
        routing = advssh.get_routing()
        self.assertEquals(routing['comment'], ['Hello', 'World', '!'])

    def test_build_ssh_config_with_multiline_localforward_onliner(self):
        contents = """
[localhost]
user = toto
localforward = 1 2.3.4.5 6 \n 7 8.9.10.11 12
port = 22
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='localhost',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        self.assertEquals(len(arr), 11)
        dest = """
# assh version: {}

Host localhost
  localforward 1 2.3.4.5 6
  localforward 7 8.9.10.11 12
  port 22
  user toto

Host *
  proxycommand assh connect %h --port=%p
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    def test_build_ssh_config_with_multiline_localforward(self):
        contents = """
[localhost]
user = toto
localforward = 1 2.3.4.5 6
               7 8.9.10.11 12
port = 22
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='localhost',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        self.assertEquals(len(arr), 11)
        dest = """
# assh version: {}

Host localhost
  localforward 1 2.3.4.5 6
  localforward 7 8.9.10.11 12
  port 22
  user toto

Host *
  proxycommand assh connect %h --port=%p
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    def test_build_ssh_config_with_multiline_comment(self):
        contents = """
[localhost]
port = 22
comment = .
          .            O
          .     _______O_
          .    /       O \\
          .   / _ _ O _ _ \\
          .    |    _    |
          .    | o | | o |
          .    |___|_|___|
          .
user = toto
"""
        set_config(contents)
        advssh = AdvancedSshConfig(hostname='localhost',
                                   configfiles=[DEFAULT_CONFIG])
        config = advssh.prepare_sshconfig()
        arr = advssh.build_sshconfig()
        string = '\n'.join(arr)
        dest = """
# assh version: {}

Host localhost
  port 22
  user toto
  # comment .
  # comment .            O
  # comment .     _______O_
  # comment .    /       O \\
  # comment .   / _ _ O _ _ \\
  # comment .    |    _    |
  # comment .    | o | | o |
  # comment .    |___|_|___|
  # comment .

Host *
  proxycommand assh connect %h --port=%p
""".format(__version__)
        self.assertEquals(string.strip(), dest.strip())

    # FIXME: test_handle_custom_proxycommand
    # FIXME: test_prepare_sshconfig_with_hostname
    # FIXME: test_routing_override_config
    # FIXME: test_connect
    # FIXME: test_dryrun
    # FIXME: test_verbose
    # FIXME: test_alias
    # FIXME: test_build_ssh_config_real_hostname



import fcntl
import os
import select
import sys
import socket
import logging


class Socket(object):
    def __init__(self, hostname, port, bufsize=4096, stdin=None, stdout=None):
        self.hostname = hostname
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bufsize = bufsize

        self.logger = logging.getLogger('assh.Socket')

        if not stdin:
            stdin = sys.stdin
        if not stdout:
            stdout = sys.stdout
        self.stdin = stdin
        self.stdout = stdout

    def run(self):
        self.socket.connect((self.hostname, self.port))
        self.communicate()

    def communicate(self):
        self.socket.setblocking(0)

        fd = sys.stdin.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        inputs = [self.socket, self.stdin]

        while True:
            read_ready, write_ready, in_error = select.select(inputs, [], [])
            for sock in read_ready:
                if sock == self.stdin:
                    buffer = self.stdin.read(self.bufsize)
                    try:
                        while buffer != '':
                            self.socket.send(buffer)
                            buffer = self.stdin.read(self.bufsize)
                    except:
                        pass
                else:
                    try:
                        buffer = self.socket.recv(self.bufsize)
                        while buffer != '':
                            self.stdout.write(buffer)
                            self.stdout.flush()
                            buffer = self.socket.recv(self.bufsize)
                        if buffer == '':
                            self.logger.warn('Server disconnected')
                            return
                    except socket.error:
                        pass



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import subprocess
import os
import sys
import re
import logging
from collections import OrderedDict
from time import sleep

from .config import Config
from .utils import (safe_makedirs, value_interpolate, construct_proxy_commands,
                    shellquotemultiple)
from . import __version__


class AdvancedSshConfig(object):

    def __init__(self, hostname=None, port=None, configfiles=None,
                 verbose=False, dry_run=False, proxy_type='nc',
                 timeout=180, use_python_socket=False,
                 ssh_config_file='~/.ssh/config', force=False):

        self.verbose, self.dry_run = verbose, dry_run
        self.hostname, self.port = hostname, port
        self.proxy_type, self.timeout = proxy_type, timeout
        self.user_python_socket = use_python_socket

        self.logger = logging.getLogger('assh.AdvancedSshConfig')

        # Initializes the Config object
        if not configfiles:
            configfiles = [
                '/etc/ssh/config.advanced',
                '~/.ssh/config.advanced',
                ]
        self.config = Config(configfiles=configfiles)
        self.ssh_config_file = ssh_config_file
        ssh_config_file_version = self.ssh_config_file_version()
        if ssh_config_file_version != __version__:
            self.logger.error(
                'ssh_config file is at version {}, but Advanced SSH config is '
                'at version {}'.format(ssh_config_file_version, __version__)
            )
        self.force = force

    @property
    def controlpath_dir(self):
        controlpath = self.config.get('controlpath',
                                      'default',
                                      '/tmp/advssh_cm/')
        directory = os.path.dirname(os.path.expanduser(controlpath))
        directory = os.path.join(directory, self.hostname)
        directory = os.path.dirname(directory)
        return directory

    def ssh_config_file_version(self, filename=None):
        if not filename:
            filename = self.ssh_config_file
        try:
            filepath = os.path.expanduser(filename)
            f = open(filepath, 'r')
            first_line = f.readline()
            if first_line.startswith('# assh version: '):
                return first_line.split(' ')[-1].strip()
            return None
        except IOError:
            return None
        return None

    def get_routing(self):
        routing = {}
        safe_makedirs(self.controlpath_dir)

        section = None
        for sect in self.config.parser.sections():
            if re.match(sect, self.hostname):
                section = sect

        self.logger.debug('section "{}" '.format(section))

        # Parse special routing
        path = self.hostname.split('/')

        args = {}
        options = {
            'p': 'Port',
            'l': 'User',
            'h': 'Hostname',
            'i': 'IdentityFile'
            }
        default_options = {
            'p': None,
            'h': path[0]
            }
        if self.port:
            default_options['p'] = self.port
        updated = False
        for key in options:
            cfval = self.config.get(options[key],
                                    path[0],
                                    default_options.get(key))
            value = value_interpolate(cfval)
            if cfval != value:
                updated = True
                self.config.parser.set(section, options[key], value)
                args[key] = value

            self.logger.debug(
                'get (-%-1s) %-12s : %s',
                key, options[key], value
            )
            if value:
                args[key] = value

        # FIXME: determine if we need to rebuild .ssh/config
        self.write_sshconfig()
        self.logger.debug('Config updated. Need to restart SSH!?')

        self.logger.debug('args: {}'.format(args))

        routing['verbose'] = self.verbose
        routing['proxy_type'] = self.proxy_type
        for special_key in ('comment', 'password', 'gateways',
                            'reallocalcommand'):
            routing[special_key] = self.config.get(special_key, path[-1], None)
        if not routing['gateways']:
            routing['gateways'] = ['direct']
        else:
            routing['gateways'] = routing['gateways'].split(' ')

        routing['gateway_route'] = path[1:]
        routing['hostname'] = args['h']
        routing['port'] = self.port
        if not routing['port'] and 'p' in args:
            routing['port'] = int(args['p'])
        if not routing['port']:
            routing['port'] = 22
        routing['proxy_commands'] = construct_proxy_commands(routing)

        self.logger.debug('Routing:')
        for key, value in routing.iteritems():
            self.logger.debug('  {0}: {1}'.format(key, value))

        return routing

    def connect(self, routing):
        for gateway in routing['gateways']:
            if gateway != 'direct':
                routing['gateway_route'] += [gateway]
                self.logger.info('Using gateway: '
                                 '{}'.format(routing['gateway_route']))
            else:
                self.logger.info('Direct connection')
            cmd = []
            if len(routing['gateway_route']):
                cmd += ['ssh', '/'.join(routing['gateway_route'])]
                cmd.append(shellquotemultiple(routing['proxy_commands']))
                self.logger.info('cmd: {}'.format(cmd))
            else:
                cmd = routing['proxy_commands'][0]

            self.logger.info('Connection command {}'.format(map(str, cmd)))

            if not self.dry_run:
                self.connect_once(routing, cmd)

    def connect_once(self, routing, cmd):
        comment = routing.get('comment', None)
        if comment:
            sys.stderr.write('{}\n'.format('\n'.join(comment)))

        password = routing.get('password', None)
        if password:
            sys.stderr.write('password: {}\n'.format(password))

        rlc_process = None
        if routing['reallocalcommand']:
            self.logger.info(
                'Executing localcommand: {}'.format(
                    routing['reallocalcommand']
                )
            )
            rlc_cmd = ['/bin/sh', '-c', routing['reallocalcommand']]
            rlc_process = subprocess.Popen(rlc_cmd,
                                           stdout=sys.stderr,
                                           stderr=sys.stderr)
            sleep(.1)

        if self.user_python_socket \
                and not len(routing['gateway_route']):
            self.logger.info('Using Python socket')
            from .network import Socket
            socket = Socket(routing['hostname'], routing['port'])
            socket.run()
        else:
            self.logger.info('Using ProxyCommand')
            proxy_process = subprocess.Popen(map(str, cmd))
            if proxy_process.wait() != 0:
                self.logger.critical('There were some errors')

        if rlc_process is not None:
            print(rlc_process)
            rlc_process.kill()

    def write_sshconfig(self, filename=None):
        if not self.force and self.ssh_config_file_version() != __version__:
            self.logger.error('Cannot save ssh_config_file, versions differ. '
                              'Use -f to force')
            return False
        if not filename:
            filename = self.ssh_config_file
        config = self.build_sshconfig()
        if self.dry_run:
            self.logger.error(
                'Without dry-run, the file {} should be replaced by the '
                'following content'.format(filename))
            self.logger.error('\n'.join(config))
        else:
            fhandle = open(os.path.expanduser(filename), 'w+')
            fhandle.write('\n'.join(config))
            fhandle.close()

    def build_sshconfig(self):
        config = []

        config.append('# assh version: {}'.format(__version__))
        config.append('')

        hosts = self.prepare_sshconfig()
        od = OrderedDict(sorted(hosts.items()))
        for entry in od.values():
            # '*' host must be at the end of the file
            if entry.host in ('*', 'default'):
                continue
            else:
                config += entry.build_sshconfig()

        # '*' host
        if 'default' in hosts:
            config += hosts['default'].build_sshconfig()

        return config

    def prepare_sshconfig(self):
        hosts = {}
        for host in self.config.full.values():
            host.resolve()
            hosts[host.host] = host
        return hosts



# -*- coding: utf-8 -*-

import unittest
import os

from advanced_ssh_config.utils import (safe_makedirs, value_interpolate,
                                       construct_proxy_commands, shellquote,
                                       shellquotemultiple)
from advanced_ssh_config.exceptions import ConfigError

from . import PREFIX


class TestContructProxyCommand(unittest.TestCase):

    def test_no_arg(self):
        self.assertRaises(TypeError, construct_proxy_commands)

    def test_empty_arg(self):
        self.assertRaises(ValueError, construct_proxy_commands, {})

    def test_minimal_valid(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'port': 42,
        })
        self.assertEqual(
            command, [['nc', '-w', 180, '-G', 5, 'aaa', 42], ['nc', 'aaa', 42]]
        )

    def test_minimal_nc(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'proxy_type': 'nc',
            'port': 42,
        })
        self.assertEqual(
            command, [['nc', '-w', 180, '-G', 5, 'aaa', 42], ['nc', 'aaa', 42]]
        )

    def test_full_nc(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'port': 42,
            'verbose': True,
            'proxy_type': 'nc',
            'timeout': 45,
        })
        self.assertEqual(
            command, [
                ['nc', '-v', '-w', 45, '-G', 5, 'aaa', 42],
                ['nc', 'aaa', 42]
            ]
        )

    def test_invalid_proxy_type(self):
        args = {
            'hostname': 'aaa',
            'port': 42,
            'proxy_type': 'fake',
            }
        self.assertRaises(ValueError, construct_proxy_commands, args)

    def test_minimal_socat(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'proxy_type': 'socat',
            'port': 42,
        })
        self.assertEqual(command, [['socat', 'STDIN', 'TCP:aaa:42']])

    def test_minimal_socat_http_proxy(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'proxy_type': 'socat_http_proxy',
            'http_proxy_host': 'bbb',
            'http_proxy_port': 43,
            'port': 42,
        })
        self.assertEqual(
            command, [['socat', 'STDIN', 'PROXY:bbb:aaa:42,proxyport=43']]
        )

    def test_minimal_socat_socks(self):
        command = construct_proxy_commands({
            'hostname': 'aaa',
            'proxy_type': 'socat_socks',
            'socks_host': 'bbb',
            'socks_port': 43,
            'port': 42,
        })
        self.assertEqual(
            command, [['socat', 'STDIN', 'SOCKS:bbb:aaa:42,socksport=43']]
        )

    # FIXME: test_custom_handler


class TestSafeMakedirs(unittest.TestCase):

    def setUp(self):
        if os.path.exists(PREFIX):
            os.system('rm -rf {}'.format(PREFIX))
        os.makedirs(PREFIX)

    def test_already_exists(self):
        safe_makedirs('{}/dir'.format(PREFIX))
        safe_makedirs('{}/dir'.format(PREFIX))

    def test_invalid(self):
        for path in ('/dev/null/test',):
            self.assertRaises(OSError, safe_makedirs, path)

    def test_makedirs_on_file(self):
        open('{}/file'.format(PREFIX), 'w').write('hello')
        self.assertRaises(OSError, safe_makedirs, '{}/file/dir'.format(PREFIX))


class TestValueInterpolate(unittest.TestCase):

    def setUp(self):
        if os.environ.get('TEST_INTERPOLATE'):
            del os.environ['TEST_INTERPOLATE']

    def test_interpolate_success(self):
        os.environ['TEST_INTERPOLATE'] = 'titi'
        self.assertEquals(value_interpolate('$TEST_INTERPOLATE'), 'titi')

    def test_interpolate_no_match(self):
        self.assertEquals(
            value_interpolate('$TEST_INTERPOLATE'), '$TEST_INTERPOLATE'
        )

    def test_interpolate_not_interpolable(self):
        os.environ['TEST_INTERPOLATE'] = 'titi'
        self.assertEquals(
            value_interpolate('TEST_INTERPOLATE'), 'TEST_INTERPOLATE'
        )

    def test_interpolate_interpolate_recursive(self):
        os.environ['TEST_INTERPOLATE'] = '$TEST_INTERPOLATE_2'
        os.environ['TEST_INTERPOLATE_2'] = '$TEST_INTERPOLATE_3'
        os.environ['TEST_INTERPOLATE_3'] = '$TEST_INTERPOLATE_4'
        os.environ['TEST_INTERPOLATE_4'] = 'tutu'
        self.assertEquals(value_interpolate('$TEST_INTERPOLATE'), 'tutu')

    def test_interpolate_interpolate_loop(self):
        os.environ['TEST_INTERPOLATE'] = '$TEST_INTERPOLATE'
        self.assertRaises(ConfigError, value_interpolate, '$TEST_INTERPOLATE')

    def test_interpolate_interpolate_loop_complex(self):
        os.environ['TEST_INTERPOLATE'] = '$TEST_INTERPOLATE_2'
        os.environ['TEST_INTERPOLATE_2'] = '$TEST_INTERPOLATE_3'
        os.environ['TEST_INTERPOLATE_3'] = '$TEST_INTERPOLATE'
        self.assertRaises(ConfigError, value_interpolate, '$TEST_INTERPOLATE')


class TestShellQuote(unittest.TestCase):

    def test_shellquote_simple(self):
        self.assertEquals(shellquote(["aaa"]), 'aaa')
        self.assertEquals(shellquote(["aaa", "bbb"]), 'aaa bbb')
        self.assertEquals(shellquote(["aaa", "bbb", 42]), 'aaa bbb 42')

    def test_shellquote_empty(self):
        self.assertEquals(shellquote([]), '')

    def test_shellquote_escape(self):
        self.assertEquals(shellquote(["test'test"]), "'test\\'test'")
        self.assertEquals(shellquote(["test\\test"]), "'test\\\\test'")
        self.assertEquals(shellquote(["test\"test"]), "test\"test")

    def test_shellquote_complex(self):
        self.assertEquals(
            shellquote(["ssh", "manfred's imac", '-p', 4242]),
            "ssh 'manfred\\'s imac' -p 4242"
        )

    def test_shellquote_not_list(self):
        self.assertRaises(ValueError, shellquote, 'aaa')


class TestShellQuoteMultiple(unittest.TestCase):

    def test_shellquote_multiple_simple(self):
        self.assertEquals(
            shellquotemultiple([['aaa', 'bbb', 42], ['ccc', 'ddd', 43]]),
            '(aaa bbb 42 2>/dev/null || ccc ddd 43)'
        )

    def test_shellquote_not_list_of_list(self):
        self.assertRaises(ValueError, shellquotemultiple, [42, 42])
        self.assertRaises(ValueError, shellquotemultiple, 42)
        self.assertRaises(ValueError, shellquotemultiple, "42")



# -*- coding: utf-8 -*-

import os

from advanced_ssh_config.config import Config


PREFIX = '/tmp/test-asc'
DEFAULT_CONFIG = os.path.join(PREFIX, 'config.advanced')


def write_config(contents, name='config.advanced'):
    with open(os.path.join(PREFIX, name), 'w') as f:
        f.write(contents)


def prepare_config():
    os.system('rm -rf {}'.format(PREFIX))
    os.makedirs(PREFIX)


def set_config(contents, load=True):
    contents = contents.format(PREFIX)
    write_config(contents)
    if load:
        config = Config([DEFAULT_CONFIG])
        return config
    return True



# -*- coding: utf-8 -*-

import re

PROXY_REGEX = re.compile(r"^(proxycommand)\s*=*\s*(.*)", re.I)


def parse_ssh_config(file_obj):
    """
    Read an OpenSSH config from the given file object.

    Small adaptation of the paramiko.config.SSH_Config.parse method
    https://github.com/paramiko/paramiko/blob/master/paramiko/config.py

    @param file_obj: a file-like object to read the config file from
    @type file_obj: file
    """
    hosts = {}
    host = {"host": ['*'], "config": {}}
    for line in file_obj:
        line = line.rstrip('\n').lstrip()
        if (line == '') or (line[0] == '#'):
            continue
        if '=' in line:
            # Ensure ProxyCommand gets properly split
            if line.lower().strip().startswith('proxycommand'):
                match = PROXY_REGEX.match(line)
                key, value = match.group(1).lower(), match.group(2)
            else:
                key, value = line.split('=', 1)
                key = key.strip().lower()
        else:
            # find first whitespace, and split there
            i = 0
            while (i < len(line)) and not line[i].isspace():
                i += 1
            if i == len(line):
                raise Exception('Unparsable line: %r' % line)
            key = line[:i].lower()
            value = line[i:].lstrip()

        if key == 'host':
            hosts[host['host'][0]] = host['config']
            value = value.split()
            host = {key: value, 'config': {}}

        # identityfile, localforward, remoteforward keys are special cases,
        # since they are allowed to be specified multiple times and they
        # should be tried in order of specification.
        elif key in ['identityfile', 'localforward', 'remoteforward']:
            if key in host['config']:
                host['config'][key].append(value)
            else:
                host['config'][key] = [value]
        elif key not in host['config']:
            host['config'].update({key: value})
    hosts[host['host'][0]] = host['config']
    return hosts



# -*- coding: utf-8 -*-

import unittest
import sys
import os

from advanced_ssh_config.bin import advanced_ssh_config_parse_options


class RedirectStdStreams(object):
    def __init__(self, stdout=None, stderr=None):
        devnull = open(os.devnull, 'w')
        if stdout == 'devnull':
            stdout = devnull
        if stderr == 'devnull':
            stderr = devnull

        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


def _advanced_ssh_config_parse_options(*args):
    sys.argv = [''] + list(map(str, args))
    return advanced_ssh_config_parse_options()


class TestParseOptions(unittest.TestCase):

    def test_with_args_without_options(self):
        self.assertRaises(
            ValueError, _advanced_ssh_config_parse_options, 'localhost'
        )

    def test_with_args_with_options(self):
        self.assertRaises(
            ValueError, _advanced_ssh_config_parse_options, '--port=22',
            'localhost'
        )
        self.assertRaises(
            ValueError, _advanced_ssh_config_parse_options, 'localhost',
            '--port=22'
        )

    def test_unkown_options(self):
        with RedirectStdStreams(stdout='devnull', stderr='devnull'):
            self.assertRaises(
                SystemExit, _advanced_ssh_config_parse_options, '--toto=titi'
            )

    def test_default(self):
        options = _advanced_ssh_config_parse_options('connect', '1.2.3.4')
        should_be = {
            'log_level': None,
            'verbose': None,
            'dry_run': None,
            'command': 'connect',
            'hostname': '1.2.3.4',
            'port': None,
            'config_file': '~/.ssh/config',
            'force': False
            }
        self.assertEqual(options, should_be)

    def test_option_valid_host(self):
        for host in ('1.2.3.4',
                     'localhost',
                     'google.com',
                     'localhost/localhost',
                     '1.2.3.4.5',
                     '::1',
                     'fe80::1%lo0'):
            options = _advanced_ssh_config_parse_options('connect', host)
            self.assertEqual(options.hostname, host)

    def test_option_invalid_host(self):
        with RedirectStdStreams(stdout='devnull', stderr='devnull'):
            self.assertRaises(
                SystemExit, _advanced_ssh_config_parse_options, 'connect'
            )
            self.assertRaises(
                SystemExit, _advanced_ssh_config_parse_options, 'info'
            )

    def test_option_valid_port(self):
        for port in (22, 1, 65535):
            options = _advanced_ssh_config_parse_options(
                'connect', 'localhost', '--port={}'.format(port)
            )
            self.assertEqual(options.port, port)

            options = _advanced_ssh_config_parse_options(
                'connect', 'localhost', '-p', port
            )
            self.assertEqual(options.port, port)

    def test_option_no_port(self):
        options = _advanced_ssh_config_parse_options('connect', 'localhost')
        self.assertEqual(options.port, None)

    def test_option_invalid_port(self):
        with RedirectStdStreams(stdout='devnull', stderr='devnull'):
            self.assertRaises(
                SystemExit, _advanced_ssh_config_parse_options,
                'connect', 'localhost', '--port'
            )
            self.assertRaises(
                SystemExit, _advanced_ssh_config_parse_options,
                'connect', 'localhost', '-p'
            )

        for port in (-1, 65536, 'test', '', '1.0'):
            self.assertRaises(
                ValueError, _advanced_ssh_config_parse_options,
                'connect', 'localhost', '--port={}'.format(port)
            )
            self.assertRaises(
                ValueError, _advanced_ssh_config_parse_options,
                'connect', 'localhost', '-p', port
            )



# -*- coding: utf-8 -*-

import sys
import optparse
import logging
import re
import os

from . import __version__
from .utils import (
    validate_host, validate_port, parent_ssh_process_info, setup_logging
    )
from .config import Config
from .exceptions import ConfigError
from .advanced_ssh_config import AdvancedSshConfig
from .ssh_config import parse_ssh_config


AVAILABLE_COMMANDS = (
    'build',
    'connect',
    'generate-etc-hosts',
    'help',
    'info',
    'init',
    'stats',
)


def advanced_ssh_config_parse_options():
    parser = optparse.OptionParser(
        usage="""%prog [OPTIONS] COMMAND [arg...]

Commands:
  build                 Build .ssh/config based on .ssh/config.advanced
  connect <host>        Open a connection to <host>
  info <host>           Print connection informations
  init                  Build a .ssh/config.advanced file based on .ssh/config
  generate-etc-hosts    Print a /etc/hosts file of .ssh/config.advanced
  stats                 Print statistics""",
        version='%prog {0}'.format(__version__)
    )
    parser.add_option('-p', '--port',
                      dest='port',
                      default=None,
                      help='SSH port')

    parser.add_option('-c', '--config',
                      dest='config_file',
                      default='~/.ssh/config',
                      help='ssh_config file')

    parser.add_option('-f', '--force',
                      dest='force',
                      default=False,
                      action='store_true')

    parser.add_option('-v', '--verbose',
                      dest='verbose',
                      action='store_true')

    parser.add_option('-l', '--log_level',
                      dest='log_level')

    parser.add_option('--dry-run',
                      action='store_true',
                      dest='dry_run')

    (options, args) = parser.parse_args()

    # Must specify a COMMAND
    if not len(args):
        parser.print_help()
        exit(-1)

    options.command = args[0]

    # COMMAND must exists
    if options.command not in AVAILABLE_COMMANDS:
        raise ValueError("'{}' is not a valid command."
                         .format(options.command))

    if options.command == 'help':
        parser.print_help()
        exit(-1)

    options.hostname = None
    # some COMMANDS needs a HOST as argument
    if options.command in ('connect', 'info'):
        if len(args) < 2:
            parser.print_help()
            exit(-1)
        options.hostname = args[1]
        validate_host(options.hostname)
        if options.port is not None:
            validate_port(options.port)
            options.port = int(options.port)

    return options


def keyboard_interrupt(fn):
    """ KeyboardInterrupt interceptor decorator. """
    def wrapper():
        logger = logging.getLogger('assh.advanced_ssh_config')
        try:
            fn()
        except KeyboardInterrupt:
            logging.fatal('Advanced SSH interrupted, bye.')
            sys.exit(1)
    return wrapper


def assh_from_options(options):
    return AdvancedSshConfig(
        hostname=options.hostname,
        port=options.port,
        verbose=options.verbose,
        dry_run=options.dry_run,
        ssh_config_file=options.config_file,
        force=options.force
    )


@keyboard_interrupt
def advanced_ssh_config():
    """ advanced-ssh-config entry-point. """
    try:
        options = advanced_ssh_config_parse_options()
    except ValueError as err:
        logging.fatal(err.message)
        sys.exit(1)

    parent = parent_ssh_process_info()
    setup_logging(options, parent)
    logger = logging.getLogger('assh.advanced_ssh_config')

    try:
        if options.command == 'build':
            ssh = assh_from_options(options)
            ssh.write_sshconfig()
            print("The file '{}' has been rebuilt".format(options.config_file))
            # FIXME: print some stats
            # FIXME: print diff

        elif options.command == 'connect':
            ssh = assh_from_options(options)
            routing = ssh.get_routing()
            ssh.connect(routing)

        elif options.command == 'info':
            ssh = assh_from_options(options)
            routing = ssh.get_routing()
            for key, value in routing.items():
                if value is None:
                    continue
                if isinstance(value, list):
                    if len(value):
                        print('{}:'.format(key))
                        print('\n'.join(['- {}'.format(line) for line in value]))
                else:
                    print('{}: {}'.format(key, value))

        elif options.command == 'generate-etc-hosts':
            assh_to_etchosts()

        elif options.command == 'stats':
            ssh = assh_from_options(options)
            hosts = ssh.prepare_sshconfig()
            print('{} hosts'.format(len(hosts.keys())))


        elif options.command == 'init':
            ssh_config_file = os.path.expanduser('~/.ssh/config')
            with open(ssh_config_file, 'r') as file_descriptor:
                config = parse_ssh_config(file_descriptor)
                for host, config in config.iteritems():

                    # Escape regex
                    if host == '*':
                        escaped_host = 'default'
                    else:
                        escaped_host = host.replace('.', '\\.')
                        escaped_host = escaped_host.replace('*', '.*')
                        # escaped_host = '^{0}$'.format(escaped_host)

                    print('[{0}]'.format(escaped_host))
                    print('# non-escaped host: {}'.format(host))
                    for key, value in config.iteritems():
                        if isinstance(value, list):
                            for entry in value:
                                print('{0} = {1}'.format(key, entry))
                            pass
                        else:
                            print('{0} = {1}'.format(key, value))
                    print('')


        else:
            raise NotImplementedError(
                "Command '{}' not yet implemented".format(options.command)
            )

    except ConfigError as err:
        sys.stderr.write(err.message)

    except Exception as err:
        sys.stderr.write(err.__str__().strip() + '\n')


def assh_to_etchosts():
    """ assh-to-etchosts entry-point. """
    print('## Automatically generated with assh-to-etchosts')
    print('')
    configfiles = [
        '/etc/ssh/config.advanced',
        '~/.ssh/config.advanced',
    ]
    config = Config(configfiles=configfiles)
    hosts = {}
    for sect in config.parser.sections():
        ip = config.get('hostname', sect)
        if ip:
            if '$' in ip:
                continue  # command
            # FIXME: handle IPV6
            if not re.match(r'[0-9\.]+', ip):
                # FIXME: try to resolve non-ip hostnames
                continue
            if ip not in hosts:
                hosts[ip] = []
            if sect not in hosts[ip]:
                hosts[ip].append(sect)
    for ip in sorted(hosts.keys()):
        hostnames = hosts[ip]
        print("{:40} {}".format(ip, ' '.join(hostnames)))



# -*- coding: utf-8 -*-

import unittest
import os

from advanced_ssh_config.config import Config
from advanced_ssh_config.exceptions import ConfigError
from . import set_config, prepare_config, write_config, PREFIX, DEFAULT_CONFIG


class TestConfig(unittest.TestCase):

    def setUp(self):
        prepare_config()

    def test_initialize_config(self):
        config = Config([DEFAULT_CONFIG])
        self.assertIsInstance(config, Config)

    def test_include_existing_files(self):
        write_config('', name='include-1')
        write_config('', name='include-2')
        contents = """
[default]
Includes = {0}/include-1 {0}/include-2
"""
        config = set_config(contents)
        self.assertEquals(config.loaded_files, [
            DEFAULT_CONFIG,
            '{}/include-1'.format(PREFIX),
            '{}/include-2'.format(PREFIX),
        ])

    def test_include_not_exists(self):
        contents = """
[default]
Includes = {0}/include-1 {0}/include-2
"""
        set_config(contents, load=False)
        self.assertRaises(ConfigError, Config, [DEFAULT_CONFIG])

    def test_include_same_file(self):
        write_config('', name='include-1')
        contents = """
[default]
Includes = {0}/include-1 {0}/include-1
"""
        config = set_config(contents)
        self.assertEquals(config.loaded_files, [
            DEFAULT_CONFIG,
            '{}/include-1'.format(PREFIX),
        ])

    def test_sections_simple(self):
        contents = """
[hosta]
[default]
"""
        config = set_config(contents)
        self.assertEquals(config.sections, ['default', 'hosta'])

    def test_sections_with_double(self):
        contents = """
[hosta]
[hosta]
[default]
"""
        config = set_config(contents)
        self.assertEquals(config.sections, ['default', 'hosta'])

    def test_sections_with_case(self):
        contents = """
[hosta]
[hostA]
[default]
"""
        config = set_config(contents)
        self.assertEquals(config.sections, ['default', 'hostA', 'hosta'])

    def test_sections_with_regex(self):
        contents = """
[hosta]
[host.*]
[default]
"""
        config = set_config(contents)
        self.assertEquals(config.sections, ['default', 'host.*', 'hosta'])

    def test_get_simple(self):
        contents = """
[hosta]
hostname = 1.2.3.4
"""
        config = set_config(contents)
        self.assertEquals(config.get('Hostname', 'hosta'), '1.2.3.4')
        self.assertEquals(config.get('hostname', 'hosta'), '1.2.3.4')

    def test_get_key_not_found(self):
        contents = """
[hosta]
"""
        config = set_config(contents)
        self.assertEquals(config.get('Hostname', 'hosta'), None)
        self.assertEquals(
            config.get('Hostname', 'hosta', 'localhost'), 'localhost'
        )

    def test_get_host_not_found(self):
        contents = """
[default]
port = 22
"""
        config = set_config(contents)
        self.assertEquals(config.get('Port', 'hosta'), '22')

    def test_get_host_and_key_not_found(self):
        config = set_config('')
        self.assertEquals(config.get('Port', 'hosta'), None)

    def test_host_wildcard(self):
        contents = """
[aaa.*]
port = 25

[.*bbb]
port = 24

[ccc.*ddd]
port = 23

[.*eee.*]
port = 22

[default]
port = 21
"""
        config = set_config(contents)
        self.assertEquals(config.get('Port', 'aaa'), '25')
        self.assertEquals(config.get('Port', 'aaa42'), '25')
        self.assertEquals(config.get('Port', '42aaa'), '21')

        self.assertEquals(config.get('Port', 'bbb'), '24')
        self.assertEquals(config.get('Port', 'bbb42'), '24')  # strange
        self.assertEquals(config.get('Port', '42bbb'), '24')

        self.assertEquals(config.get('Port', 'cccddd'), '23')
        self.assertEquals(config.get('Port', 'ccc42ddd'), '23')

        self.assertEquals(config.get('Port', 'eee'), '22')
        self.assertEquals(config.get('Port', '42eee'), '22')
        self.assertEquals(config.get('Port', 'eee42'), '22')
        self.assertEquals(config.get('Port', '42eee42'), '22')

    def test_host_invalid_wildcard(self):
        contents = """
[aaa.+]
port = 25
"""
        set_config(contents, load=False)
        self.assertRaises(ConfigError, Config, [DEFAULT_CONFIG])

    def test_multiple_line(self):
        contents = """
[test]
localforward = 1 test 2 \n 2 test 3
"""
        config = set_config(contents)
        self.assertEquals(
            config.get('localforward', 'test'), ['1 test 2', '2 test 3']
        )

    def test_one_line_list(self):
        contents = """
[test]
localforward = 1 test 2
"""
        config = set_config(contents)
        self.assertEquals(config.get('localforward', 'test'), ['1 test 2'])



# -*- coding: utf-8 -*-

import logging
import re
import os
import errno
import subprocess

from .exceptions import ConfigError


def shellquote_arg(before):
    before = str(before)
    after = before
    after = after.replace('\\', '\\\\')
    after = after.replace('\'', '\\\'')
    if after != before:
        return "'{}'".format(after)
    else:
        return before


def shellquote(cmd):
    if type(cmd) != list:
        raise ValueError('`cmd` must be a list')
    return ' '.join(map(shellquote_arg, cmd))


def shellquotemultiple(cmds):
    if type(cmds) != list:
        raise ValueError('`cmds` must be a list of list')
    for cmd in cmds:
        if type(cmd) != list:
            raise ValueError('`cmd` in `cmds` must be lists')
    if len(cmds) > 1:
        return '({})'.format(' 2>/dev/null || '.join(map(shellquote, cmds)))
    else:
        return cmds[0]


def construct_proxy_commands(config):
    cmds = []
    proxy_type = config.get('proxy_type', 'nc')
    verbose = config.get('verbose', False)
    timeout = config.get('timeout', 180)
    connection_timeout = config.get('connection_timeout', 5)
    if 'hostname' not in config or 'port' not in config:
        raise ValueError('hostname and port must be configured')
    hostname, port = config['hostname'], config['port']
    if proxy_type in ('nc', 'ncat', 'netcat'):
        cmd = []  # cmd with options
        cmd.append(proxy_type)
        if verbose:
            cmd.append('-v')
        if timeout:
            cmd.append('-w')
            cmd.append(timeout)
        if connection_timeout:
            cmd.append('-G')
            cmd.append(connection_timeout)
        cmd.append(hostname)
        cmd.append(port)
        cmds.append(cmd)
        cmd = []  # cmd without options
        cmd.append(proxy_type)
        cmd.append(hostname)
        cmd.append(port)
        cmds.append(cmd)
    elif proxy_type == 'socat':
        cmd = []
        cmd.append('socat')
        cmd.append('STDIN')
        cmd.append('TCP:{}:{}'.format(hostname, port))
        cmds.append(cmd)
    elif proxy_type == 'socat_http_proxy':
        cmd = []
        cmd.append('socat')
        cmd.append('STDIN')
        args = [
            config.get('http_proxy_host', '127.0.0.1'),
            hostname,
            port,
            config.get('http_proxy_port', 3128),
            ]
        if config.get('http_proxy_auth', None):
            args.append(config.get('http_proxy_auth'))
            cmd.append('PROXY:{}:{}:{},'
                       'proxyport={},'
                       'proxyauth={}'.format(*args))
        else:
            cmd.append('PROXY:{}:{}:{},proxyport={}'.format(*args))
        cmds.append(cmd)
    elif proxy_type == 'socat_socks':
        cmd = []
        cmd.append('socat')
        cmd.append('STDIN')
        args = [
            config.get('socks_host', '127.0.0.1'),
            hostname,
            port,
            config.get('socks_port', 1080),
            ]
        cmd.append('SOCKS:{}:{}:{},socksport={}'.format(*args))
        cmds.append(cmd)
    else:
        raise ValueError('proxy_type `{}` is not handled'.format(proxy_type))
    return cmds


def validate_host(host):
    if not type(host).__name__ in ('str', 'unicode'):
        raise ValueError('host must be a string')
    if len(host) == 0:
        raise ValueError('host is empty')


def validate_port(port):
    if type(port).__name__ == 'str':
        try:
            port = int(port)
        except ValueError:
            raise ValueError('port must be a number')
    if type(port).__name__ != 'int':
        raise ValueError('port must be an integer')
    if port < 1 or port > 65535:
        raise ValueError('port must be between 1-65535')


def safe_makedirs(directory):
    try:
        os.makedirs(directory)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise exception


def value_interpolate(value, already_interpolated=None):
    if not already_interpolated:
        already_interpolated = []
    if type(value) == int:
        return value

    # Variable
    matches = value and re.match(r'\$(\w+)', value) or None
    if matches:
        var = matches.group(1)
        if var in already_interpolated:
            raise ConfigError('Interpolation loop')
        val = os.environ.get(var)
        if val:
            logger = logging.getLogger('assh.value_interpolate')
            logger.debug('\'{}\' => \'{}\''.format(value, val))
            new_value = re.sub(r'\${}'.format(var), val, value)
            return value_interpolate(new_value, already_interpolated + [var])

    # Bash command
    if type(value) == str and value.lstrip()[:2] == '$(' and value.rstrip()[-1] == ')':
        logger = logging.getLogger('assh.value_interpolate')
        command = value.strip()[2:-1]
        logger.info('Executing "{}"'.format(command))
        ret = subprocess.check_output(['/bin/sh', '-c', command]).strip()
        logger.info('Result: {}'.format(ret))
        return ret

    return value


def parent_ssh_process_info():
    try:
        import psutil
        return psutil.Process(os.getppid())
    except:
        return None


def setup_logging(options, parent=None):

    def get_logging_level():
        if options.verbose:
            return logging.DEBUG

        if parent:
            parent_level = 0
            for arg in parent.cmdline:
                if arg == '-v':
                    parent_level += 1
                elif arg == '-vv':
                    parent_level += 2
                elif arg == '-vvv':
                    parent_level += 3
            if parent_level:
                parent_level = min(parent_level + 2, 4)
                return LOGGING_LEVELS.get(parent_level, logging.INFO)

        environ_log_level = os.environ.get('ASSH_LOG_LEVEL', None)
        if environ_log_level:
            return LOGGING_LEVELS.get(environ_log_level, logging.ERROR)

        return LOGGING_LEVELS.get(options.log_level, logging.ERROR)

    logger = logging.getLogger('assh')
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.setLevel(get_logging_level())
    return logger


LOGGING_LEVELS = {
    'crit':     logging.CRITICAL,
    'critical': logging.CRITICAL,
    0:          logging.CRITICAL,
    'err':      logging.ERROR,
    'error':    logging.ERROR,
    1:          logging.ERROR,
    'warn':     logging.WARNING,
    'warning':  logging.WARNING,
    2:          logging.WARNING,
    'info':     logging.INFO,
    3:          logging.INFO,
    'debug':    logging.DEBUG,
    4:          logging.DEBUG,
    }



# -*- coding: utf-8 -*-

import os
import logging
import ConfigParser
import re
from collections import OrderedDict

from .exceptions import ConfigError


class ConfigHost(object):
    special_keys = (
        'hostname',
        'gateways',
        'reallocalcommand',
        'remotecommand',
        'includes',
        'inherits',
        'comment',
        'password',
        # 'proxycommand',
        )

    key_translation = {
        'alias': 'hostname',
        }

    def __init__(self, c, host, config=None, extra=None, inherited_config=None,
                 inherited_extra=None):
        self.c = c
        self.host = host
        self.config = config or {}
        self.extra = extra or {}
        self.inherited = None
        self.resolved = False

    @classmethod
    def prepare_hostname(cls, host):
        host = re.sub(r'\.\*', '*', host)
        host = re.sub(r'\\\.', '.', host)
        return host

    @classmethod
    def from_config_file(cls, c, host, entry):
        config = []
        extra_config = []
        for key, value in entry:
            if key in ConfigHost.key_translation:
                key = ConfigHost.key_translation.get(key)
            values = value.split('\n')
            values = map(str.strip, values)
            for line in values:
                # if line matches $(...), we need to put the line in extras
                if line.lstrip().find('$(') == 0 and line.rstrip()[-1] == ')':
                    extra_config.append((key, line))
                else:
                    if key in ConfigHost.special_keys:
                        extra_config.append((key, line))
                    else:
                        config.append((key, line))
        return cls(c, host, config=config, extra=extra_config)

    def config_keys(self):
        return [entry[0] for entry in self.config]

    @property
    def config_dict(self):
        if not self.resolved:
            self.resolve()
        config = {}
        for entry in self.config:
            if entry[0] in config:
                config[entry[0]].append(entry[1])
            else:
                config[entry[0]] = [entry[1]]
        return config

    @property
    def extra_dict(self, sort=True):
        extra = {}
        for entry in self.extra:
            if entry[0] in extra:
                extra[entry[0]].append(entry[1])
            else:
                extra[entry[0]] = [entry[1]]
        if sort:
            return OrderedDict(sorted(extra.items()))
        return extra

    @property
    def clean_config(self, sort=True):
        config = self.config_dict
        if self.inherited:
            config = dict(self.inherited.items() + config.items())

        # Ensure 'default' Host has a proxycommand
        if self.host == 'default' and not 'proxycommand' in config:
            config['proxycommand'] = ['assh connect %h --port=%p']

        if sort:
            return OrderedDict(sorted(config.items()))
        return config

    def resolve(self, rec=10):
        if not rec:
            raise ConfigError('Maximum recursion deptch exceeded')
        if self.resolved:
            return
        for key, value in self.extra:
            if key == 'inherits':
                if value in self.c.full:
                    parent = self.c.full[value]
                    parent.resolve(rec - 1)
                    self.inherited = parent.clean_config
                else:
                    raise ConfigError('Inheriting an unkonwn '
                                      'host: `{}`'.format(value))
        self.resolved = True

    def get_prep_value(self):
        return {
            'config': self.config,
            'extra': self.extra,
            'inherited': self.inherited,
            }

    def __repr__(self):
        max_len = 50
        dict_string = ', '.join([
            '%s=%s' % (key, str(val)[:max_len - 2] + '..'
                       if len(str(val)) > max_len else str(val))
            for key, val in self.get_prep_value().items()
            ])
        return '<{}.{} at {} - {}>'.format(self.__class__.__module__,
                                           self.__class__.__name__,
                                           hex(id(self)),
                                           dict_string)

    def build_sshconfig(self):
        sub_config = []

        if self.host == 'default':
            host = '*'
        else:
            host = self.host

        sub_config.append('Host {}'.format(host))

        attrs = OrderedDict(sorted(self.clean_config.items()))
        for key, values in attrs.iteritems():
            for value in values:
                sub_config.append('  {} {}'.format(key, value))

        for key, values in self.extra_dict.iteritems():
            for value in values:
                sub_config.append('  # {} {}'.format(key, value))

        sub_config.append('')
        return sub_config


class Config(object):
    def __init__(self, configfiles):

        self.full_cache = None
        self.configfiles = map(os.path.expanduser, configfiles)
        self.loaded_files = []

        self.logger = logging.getLogger('assh.Config')

        self.parser = ConfigParser.ConfigParser()
        self.parser.SECTCRE = re.compile(
            r'\['
            r'(?P<header>.+)'
            r'\]'
            )
        self._read()

        for section in self.sections:
            if re.sub(r'[^a-zA-Z0-9\\\.\*_-]', '', section) != section:
                raise ConfigError('Invalid characters used in '
                                  'section {}'.format(section))

    def debug(self, string=None):
        self.logger.debug(string and string or '')

    def _load_file(self, filename):
        if filename in self.loaded_files:
            return
        self.parser.read(filename)
        self.loaded_files.append(filename)

    def _read(self):
        for configfile in self.configfiles:
            self._load_file(configfile)

        # Load sub files
        includes = str(self.get('includes', 'default', '')).strip()
        for include in includes.split():
            incpath = os.path.expanduser(include)
            if os.path.exists(incpath):
                self._load_file(incpath)
            else:
                raise ConfigError('\'{}\' include not found'.format(incpath))

    @property
    def sections(self):
        return sorted(list(set(self.parser.sections() + ['default'])))

    def get_in_section(self, section, key, raw=False, vardct=None):
        if not self.parser.has_option(section, key):
            return False
        var = self.parser.get(section, key, raw, vardct)
        if key in ('identityfile', 'localforward', 'remoteforward', 'comment'):
            var = var.split('\n')
            var = map(str.strip, var)
        return var

    def get(self, key, host, default=None, vardct=None):
        for section in self.sections:
            if re.match(section, host):
                val = self.get_in_section(section, key, vardct=vardct)
                if val:
                    return val
        val = self.get_in_section('default', key)
        return val or default

    @property
    def full(self):
        if not self.full_cache:
            self.full_cache = {}
            for section in self.parser.sections():
                host = ConfigHost.prepare_hostname(section)
                config_file_entry = self.parser.items(section,
                                                      False,
                                                      {})
                conf = ConfigHost.from_config_file(self, host,
                                                   config_file_entry)
                self.full_cache[host] = conf

            if not 'default' in self.full_cache:
                self.full_cache['default'] = ConfigHost(self, 'default')

        return self.full_cache
